# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2017-08-09
#

import re
import traceback
import click
import collections


TRACE_SPLITTER = "*" * 50


def format_duration(time):
    """format duration, unit of input time should be 'second'."""

    fmt = ""
    if time < 1:
        fmt = "{:.3f} ms".format(time * 1000)
    elif time >= 1 and time < 60:
        # seconds
        fmt = "{:.3f} sec".format(time)
    elif time >= 60 and time < 3600:
        # minutes and seconds
        min = int(time / 60)
        sec = time - min * 60
        fmt = "{:d} min {:.3f} sec".format(min, sec)
    elif time >= 3600 and time < 86400:
        # hours, minutes and seconds
        hour = int(time / 3600)
        min = int((time - hour * 3600) / 60)
        sec = time - hour * 3600 - min * 60
        fmt = "{:d} hour {:d} min {:.3f} sec".format(hour, min, sec)
    else:
        # days, hours, minutes and seconds
        day = int(time / 86400)
        hour = int((time - day * 86400) / 3600)
        min = int((time - day * 86400 - hour * 3600) / 60)
        sec = time - day * 86400 - hour * 3600 - min * 60
        fmt = "{:d} day {:d} hour {:d} min {:.3f} sec".format(day, hour,
                                                              min, sec)
    return fmt


def format_traceback():
    """format traceback message."""

    return "{splitter}\n{traceback}\n{splitter}\n".format(
        splitter=TRACE_SPLITTER, traceback=traceback.format_exc())


def upperfirst(value):
    """upper first word of string, and does not change the rest case,
    such as:
        foobAr => FoobAr
    """

    if len(value) > 1:
        return value[0].upper() + value[1:]
    else:
        return value[0].upper()


def camelize(value):
    """convert string to camelcase, will split string by '_' and
    capitalize the followed word, such as:
        foo_bar => FooBar
    """

    return upperfirst(re.sub(r'(?!^)_([a-zA-Z])',
                      lambda m: m.group(1).upper(), value))


def underscore(value):
    """convert string to underscore case, will split string by capitalized
    word and replaced by '_' + lowercase, such as:
        FooBar => foo_bar
    """

    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', value)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def flatten(l):
    """make list flatten"""

    for el in l:
        try:
            is_basestring = isinstance(el, basestring)
        except NameError:
            is_basestring = isinstance(el, str)

        if isinstance(el, collections.Iterable) and not is_basestring:
            for sub in flatten(el):
                yield sub
        else:
            yield el


def convert_newline(value):
    """convert CR-LF to unix-like new line as LF"""

    return re.sub(r"\r\n", r"\n", value)


def info(message):
    """print normal message in stdout."""

    click.echo(message)


def stage(message):
    """print stage message in stdout."""

    click.secho(message, fg="green")


def warning(message):
    """print warning message in stdout."""

    click.secho("[WARNING] " + message, fg="yellow")


def error(message):
    """print error message in stdout."""

    click.secho("[ERROR] " + message, fg="red")

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2019-01-29
#


import os
from mako.template import Template
from .globals import ctx, BLOCK_SPLITTER
from .helpers import *


class Reporter(object):
    """print report to file according to output type, currently support
    'text', 'xml', 'html'.
    """

    def __init__(self):
        super(Reporter, self).__init__()

        self.output_path = None

    def report(self):
        """print report to file according to output type."""

        try:
            self.summary = (
                'mission completed in %s\n'
                'totally %d cases were executed with:\n'
                '    %d warning, %d error, %d exception\n'
                '    %d success, %d failure' % (
                    format_duration(ctx.counter.get_duration_of_app()),
                    ctx.counter.length(),
                    ctx.counter.total_warning(),
                    ctx.counter.total_error(),
                    ctx.counter.total_exception(),
                    ctx.counter.total_success(),
                    ctx.counter.total_failure()))

            if ctx.app.is_cli_mode():
                self._print_to_stdout(ctx.counter)

            if ctx.app.output in ("text", "xml", "html") \
               and ctx.output_path is not None:

                self.output_path = getattr(
                    self, "_print_to_%s" % ctx.app.output)(
                    ctx.output_path, ctx.counter)

                stage("\nreport in:\n%s" % self.output_path)
        except Exception:
            error("error occured in reporting:\n%s" % format_traceback())
        finally:
            return self.output_path

# protected
    def _print_to_stdout(self, counter):
        """print summary to stdout"""

        stage('\n\n%s' % BLOCK_SPLITTER)
        stage(self.summary)

    def _print_to_text(self, path, counter):
        """print report to text file"""

        file_path = os.path.join(path, "report.txt")

        content = Template(
            filename=os.path.join(ctx.minimo_root_path,
                                  "templates",
                                  "reports",
                                  "report.txt.mako")
        ).render_unicode(
            suite_name=ctx.suite_name,
            summary=self.summary,
            counter=counter)

        with open(file_path, "w") as f:
            f.write(convert_newline(content))

        return file_path

    def _print_to_xml(self, path, counter):
        """print report to xml file"""

        file_path = os.path.join(path, "report.xml")

        content = Template(
            filename=os.path.join(ctx.minimo_root_path,
                                  "templates",
                                  "reports",
                                  "report.xml.mako"),
        ).render_unicode(
            suite_name=ctx.suite_name,
            counter=counter)

        with open(file_path, "w") as f:
            f.write(convert_newline(content))

        return file_path

    def _print_to_html(self, path, counter):
        """print report to html file"""

        file_path = os.path.join(path, "report.html")

        content = Template(
            filename=os.path.join(ctx.minimo_root_path,
                                  "templates",
                                  "reports",
                                  "report.html.mako"),
        ).render_unicode(
            suite_name=ctx.suite_name,
            counter=counter)

        with open(file_path, "w") as f:
            f.write(convert_newline(content))

        return file_path

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2017-08-08
#


import os
import sys
import logging
from .globals import ctx


SUCCESS = logging.INFO + 1
logging.addLevelName(SUCCESS, "SUCCESS")
FAILURE = logging.ERROR + 1
logging.addLevelName(FAILURE, "FAILURE")

_LVL_PREFIX = {
    logging.INFO: "",
    logging.ERROR: "[ERROR] ",
    logging.WARNING: "[WARNING] ",
    SUCCESS: "[SUCCESS] ",
    FAILURE: "[FAILURE] ",
}


class MoFilter(logging.Filter):
    """MO filter to map customized level."""

    def filter(self, record):
        record.lvl = _LVL_PREFIX[record.levelno]
        return True


class Logger(object):
    """print log to log file and stdout.

    log directory will be placed under
    project.root/logs. by default, it will create two kinds of log file which
    are fulltrace and errors only.

    :param case: task case name, used as log file prefix.
    :param output_path: project output path, log file and report file will be
                        created under this path, normally it's `app_root/log`.
    :param max_flush_count: threshold to flush message to log file,
                            set to 10 by default.

    usage::
        logger = Logger(
            name="case name",
            output_path="path/to/project/output_path"
        )

        logger.info("information")
        logger.warning("warning message")
        logger.error("error message")

        logger.close()
    """

    def __init__(
        self,
        name="my_case",
        output_path=".",
        max_flush_count=10,
        stdout=True,
        outputs={
            "fulltrace": logging.INFO,
            "error": logging.ERROR,
        }
    ):
        super(Logger, self).__init__()

        self.name = name
        self.output_path = output_path
        self.stdout = stdout
        self.outputs = outputs
        self.max_flush_count = max_flush_count

        self._closed = True
        self._flush_count = 0

        # store file handler as:
        # {
        #    name: handler
        # }
        self._filehandlers = {}

        # setup logger
        self.open()

    def __del__(self):
        """destructor"""

        self.close()

    def open(self):
        """open log system and print log to the specified outputs"""

        if self._closed:
            # create logger
            self._logger = logging.getLogger("log-{0}".format(self.name))
            self._logger.setLevel(logging.INFO)

            # create stdout/file handler and set level
            self._logger.addFilter(MoFilter())
            formatter = \
                logging.Formatter("[%(asctime)s] %(lvl)s%(message)s")

            # add stdout handlers to logger
            if self.stdout:
                stdout = logging.StreamHandler(sys.stdout)
                stdout.setLevel(logging.NOTSET)
                stdout.setFormatter(formatter)
                self._logger.addHandler(stdout)

            # add file handlers to logger
            if len(self.outputs) > 0:
                basename = os.path.join(self.output_path, self.name)

                # check out dirs
                if not os.path.exists(self.output_path):
                    try:
                        # may conflict in concorrence running type
                        os.makedirs(self.output_path)
                    except Exception:
                        pass

                for name, level in self.outputs.items():
                    handler = logging.FileHandler(
                        "{0}.{1}.log".format(basename, name),
                        encoding="utf-8")
                    handler.setLevel(level)
                    handler.setFormatter(formatter)
                    self._filehandlers[name] = handler
                    self._logger.addHandler(handler)

            self._closed = False
        return self

    def close(self):
        """manually close log system, flush all messages to log file,
        and clear message counters.

        when the logging module is imported, it registers logging.shutdown()
        as an exit handler, that will perform an orderly shutdown by flushing
        and closing all handlers. here we provide this function to release
        resources manually if necessary in the whole working flow.
        """

        if not self._closed:
            try:
                self._flush_count = 0
                for handler in self._filehandlers.values():
                    # handler calls flush() in close()
                    handler.close()
                    self._logger.removeHandler(handler)
            except Exception:
                # do nothing
                pass

            # reset flags
            self._closed = True

    def info(self, message):
        """print normal information"""

        self.write(message, logging.INFO)

    def error(self, message):
        """print error message"""

        ctx.counter.append_error(self.name, message)
        self.write(message, logging.ERROR)

    def warning(self, message):
        """print warning message"""

        ctx.counter.append_warning(self.name, message)
        self.write(message, logging.WARNING)

    def success(self, message):
        """print task success message"""

        ctx.counter.append_success(self.name, message)
        self.write(message, SUCCESS)

    def fail(self, message):
        """print task failure message"""

        ctx.counter.append_failure(self.name, message)
        self.write(message, FAILURE)

    def write(self, message, level=logging.INFO):
        """print message to log handler according to logging level.

        message format looks like:
            '[timestamp] level-prefix message'
        """

        if not self._closed:
            self._logger.log(level, message)
            self._flush_count += 1
            if self._flush_count >= self.max_flush_count:
                self.flush()

    def flush(self):
        """flush messages to handlers"""

        for handler in self._filehandlers.values():
            handler.flush()
        self._flush_count = 0

    def get_filehandler(self, name):
        """get file handler by the specified name, if there's no such name,
        returns None.
        """

        return self._filehandlers.get(name)


class StdoutToFileLogger(Logger):
    """redirect stdout to log file"""

    def __init__(
        self,
        name="my_case",
        output_path=".",
        outputs={
            "fulltrace": logging.INFO
        }
    ):
        super(StdoutToFileLogger, self).__init__(
            name=name,
            stdout=False,
            output_path=output_path,
            outputs=outputs
        )

    def write(self, message, level=logging.INFO):
        """override write function to handle byte type message"""

        if message is None:
            _msg = ""
        else:
            _msg = message.decode("utf-8")
        self._logger.log(level, _msg)

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2018-02-26
#


import os
import time
import click
import shutil
from mako.template import Template
from ..helpers import *
from ..globals import ctx
from ..filters import convert_newline
from minimo import __version__


@click.command("init")
@click.argument("name", nargs=1)
@click.option("-t", "--template", default=None,
              help=("specify project template, optional, default is 'task'"))
@click.option("-o", "--output", default=None,
              help=("generate project to the specified path"))
def init_project(name, template=None, output=None):
    """create new project from the specified template.

    usage in cli mode:

        $ mmo init [project-name] [-t template-name-or-path] [-o output-path]

    the project will be created under 'output-path', if no 'output-path'
    specified, that will be the current working directory. if not specified
    template, minimo will initialize the project with 'task' template.
    currenty template name only supports 'task', or you can specify a path
    which contains the template.

    tip: can use 'mmo' or 'minimo' as the main command after v0.4.0.

    ----------

    usage in api mode:

        import minimo

        mmo = minimo.Application(
                    interface="api")

        # return True or False for `init` result
        result = mmo.call(
                        "init",
                        name="helloKitty",
                        output="./myprojects")
    """

    try:
        result = False

        project_name = camelize(name.replace("-", "_"))
        project_dir_name = underscore(project_name)

        if output is None:
            project_root_dir = os.getcwd()
        else:
            project_root_dir = os.path.abspath(output)
            if not os.path.exists(project_root_dir):
                os.makedirs(project_root_dir)

        project_dir = os.path.join(project_root_dir, project_dir_name)

        config = {
            "project_name": project_name,
            "date": time.strftime("%Y-%m-%d"),
            "version": __version__
        }

        # check out target path
        if os.path.exists(project_dir):
            error("directory '%s' already exsited" % project_dir_name)
        else:
            # check out template path
            template_dir = None
            if template:
                user_template_path = os.path.abspath(template)
                minimo_named_template_path = os.path.join(
                    ctx.minimo_root_path, "templates", "projects", template)

                if os.path.exists(user_template_path):
                    # user specified template path
                    template_dir = user_template_path
                elif os.path.exists(minimo_named_template_path):
                    # minimo provides template name
                    template_dir = minimo_named_template_path
                else:
                    # unrecognized template name
                    error("unrecognized project template: '%s'" % template)
            else:
                # use minimo default template
                template_dir = os.path.join(
                    ctx.minimo_root_path, "templates", "projects", "task")

            if template_dir is not None:
                info("create directory: %s" % project_dir_name)
                os.makedirs(project_dir)

                copy_template_folder(project_dir, template_dir, ".mot", config)

                result = True
    except Exception:
        result = False
    finally:
        return result


@click.command("new")
@click.argument("cases", nargs=-1)
@click.option("--author", "-a", nargs=1, type=click.STRING)
def create_new_cases(cases, author=None):
    """generate case from templates.

    usage in cli mode:

        $ mmo new [cases...] [-a author]

    for example:

        $ mmo new suite1/case1 suite2/case2 case3 [-a hellokitty]

    minimo will walk through the sub-directory of task suite, if templates
    exists in task suite, it initializes the case by the suite specified
    templates, otherwise, by the project default templates.

    if specified author name, it will be filled in the template file, or minimo
    will get the current system user as the author name.

    tip: can use 'mmo' or 'minimo' as the main command after v0.4.0.

    ----------

    usage in api mode:

        import minimo

        mmo = minimo.Application(
                    interface="api",
                    root_path=instance_project_path)

        # return successfully created cases list
        cases = mmo.call(
                    "new",
                    cases=["case1", "suite2/case1", "suite2/case2"])

    """

    try:
        success_cases = []

        if ctx.app.inst_path is None:
            error('not in minimo project root folder')
            return success_cases

        stage("prepare to create case...")

        if author is None:
            import getpass
            author = getpass.getuser()

        config = {
            "author": author,
            "date": time.strftime("%Y-%m-%d")
        }

        for case in set(cases):
            # checking templates
            dirs = ["cases"] + case.split("/")
            template_dir = None
            while len(dirs) > 0:
                dirs.pop()
                _templatedir = os.path.join(ctx.app.inst_path,
                                            *(dirs + ["templates"]))
                if os.path.exists(_templatedir):
                    template_dir = _templatedir
                    break

            if template_dir is None:
                warning(
                    "no template found, abort creating task under cases/%s" %
                    case)
            else:
                # checking target path
                target = os.path.join(ctx.app.inst_path, "cases", case)
                if os.path.exists(target):
                    warning(
                        "directory cases/%s already existed, skip this step!" %
                        case)
                    continue
                else:
                    info("create directory: cases/%s" % case)
                    os.makedirs(target)

                info("create case by project template %s" % (
                    template_dir.replace(ctx.app.inst_path,
                                         "%s.root" % ctx.app.name)))

                # copy files
                config["case_name"] = os.path.basename(case)
                copy_template_folder(target, template_dir, ".mako", config)
                success_cases.append(case)

                stage("case created under %s.root/cases" % ctx.app.name)
    except Exception:
        pass
    finally:
        return success_cases


def copy_template_file(
    dest,
    src,
    config={}
):
    """copy template file from src to dest, replace the placeholder in template
    file by the given config keywords."""

    try:
        if os.path.exists(dest):
            warning("file already existed, skip this step.")
            return False

        content = Template(
            filename=src,
            default_filters=["trim"],
        ).render_unicode(**config)

        with open(dest, "w") as f:
            f.write(convert_newline(content))

        info("\tcreate file: %s" % os.path.basename(dest))

        return True
    except Exception:
        error("fail to create file %s!\nreason:\n%s" % (
            os.path.basename(dest),
            format_traceback()))
        return False


def copy_template_folder(
    dest_dir,
    template_dir,
    template_file_suffix,
    config
):
    """copy template directory to dest."""

    for dirpath, dirs, files in os.walk(template_dir):
        dest_subdir_name = os.path.relpath(dirpath, template_dir)
        dest_subdir = os.path.join(dest_dir, dest_subdir_name)

        if "." != dest_subdir_name:
            if os.path.exists(dest_subdir):
                info("directory %s already exsited" % dest_subdir_name)
            else:
                info("create directory: %s" % dest_subdir_name)
                os.makedirs(dest_subdir)

        for file in [f for f in files if _files_filter(f)]:
            src = os.path.join(dirpath, file)
            if os.path.splitext(file)[-1] == template_file_suffix:
                # copy template file
                fparts = file.rpartition(template_file_suffix)
                dst = os.path.join(dest_subdir, "".join(
                    [fparts[0], fparts[2]]))
                copy_template_file(dst, src, config)
            else:
                # copy raw file
                dst = os.path.join(dest_subdir, file)
                shutil.copyfile(src, dst)
                info("\tcreate file: %s" % os.path.basename(dst))


def _files_filter(name):
    """filter file names out."""

    return not name.lower().endswith((
        ".pyc",
        ".pyo",
        ".placeholder"
    ))

# end




import os
import minimo
import traceback
from mako.template import Template


TRACE_SPLITTER = "*" * 50


mmo = minimo.Application(interface="api")

# return True or False for `init` result
result = mmo.call("init", name="helloKitty")


# # inst_path = os.path.dirname(__file__)
# inst_path = "/home/philip1134/projects/test/mmo-test/hello_kitty"
# mmo = minimo.Application(
#     interface="api",
#     root_path=inst_path)


# mmo.call("ls")
# # print mmo.call("version")
# # print mmo.call("init", name="hello_kitty")
# print mmo.call("new", cases=["hello", "shithappens/init", "shithappens/delete"])
# # end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2018-08-18
#


import click


class InterfaceFactory:
    """factory to create interface"""

    @classmethod
    def get(cls, type, **attrs):
        """get interface instance by type"""

        if "api" == type:
            return MoApi(**attrs)
        else:
            return click.Group(**attrs)


class MoApi(object):
    """Interface for api type"""

    def __init__(self, **attrs):
        super(MoApi, self).__init__()

        # initialize command mapper
        self.commands = attrs.get("commands") or {}

    def main(self, prog_name, **kwargs):
        return self.commands[prog_name](**kwargs)

    def add_command(self, cmd, name=None):
        """register command name and handler to command mapper."""

        name = name or cmd.name
        if name is None:
            raise TypeError('Command has no name.')
        self.commands[name] = cmd.callback

    def get_command(self, ctx, name):
        return self.commands.get(name)

    def list_commands(self, ctx):
        return sorted(self.commands)

# end



# -*- coding:utf-8 -*-

# import your extensions here, and insert them into __autoload__, minimo
# will auto load the extended cli.

# from .my-extension import my-cli

__autoload__ = (
    # my-cli
)

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2017-08-10
#


import os
from .attribute_dict import AttributeDict
from .counter import Counter


# constants
BLOCK_SPLITTER = "=" * 50
SECTION_SPLITTER = "-" * 50


# runtime context
class RuntimeContext(AttributeDict):
    """stores runtime context, includes application instance, callbacks,
    test suite name, minimo root path, and etc.
    """

    def __init__(self):
        super(RuntimeContext, self).__init__({
            "app": None,
            "counter": None,
            "reporter": None,
            "callbacks": None,
            "suite_name": None,
            "output_path": None,
            "config": AttributeDict(),
        })


ctx = RuntimeContext()
ctx.counter = Counter()
ctx.minimo_root_path = os.path.dirname(__file__)

# end



# -*- coding:utf-8 -*-


__version__ = "0.7.0"


import os
import runpy
from .app import Application
from .performer import Performer
from .callbacks import before_action, action_step, after_action
from .helpers import *
from .attribute_dict import AttributeDict
from .logger import Logger
from .ext import copy_template_file, copy_template_folder


def main():
    bin_path = os.path.join(os.getcwd(), "bin", "minimo")
    if os.path.exists(bin_path):
        runpy.run_path(bin_path)
    else:
        Application().main()

# end



# -*- coding:utf-8 -*-


from .generator import init_project, create_new_cases, \
    copy_template_file, copy_template_folder
from .commands import print_version, list_cases, run_suite


# auto load to Application cli from the following tuple
__autoload__ = (
    # generators
    init_project, create_new_cases,

    # commands
    print_version, list_cases, run_suite
)



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2019-01-29
#


import collections
from .timer import Timer


class Counter(collections.OrderedDict):
    """counter to count error/warning/exception/success/failure
    in the task process.
    """

    ERROR = "error"
    WARNING = "warning"
    EXCEPTION = "exception"
    SUCCESS = "success"
    FAILURE = "failure"
    TIMER = "timer"

    MINIMO = "__minimo_application__"

    def __init__(self):
        super(Counter, self).__init__()

        # count application informations respectively
        self._app = {
            self.ERROR: None,
            self.EXCEPTION: None,
            self.WARNING: None,
            self.SUCCESS: None,
            self.FAILURE: None,
            self.TIMER: None
        }

    def length(self):
        """get map lenght"""

        return len(self)

    def keys(self):
        """override keys() due to different odict keys()/values() type between
        python2 and python3."""

        return list(super(Counter, self).keys())

    def values(self):
        """override values() due to different odict keys()/values() type
        between python2 and python3."""

        return list(super(Counter, self).values())

    def get(self, key, flag):
        """get key's values"""

        if key in self:
            return self[key].get(flag)
        else:
            return None

    def append(self, key, flag, value):
        """append value to the key"""

        if key not in self:
            self[key] = self.__get_initialized_data()

        self[key][flag].append(value)

    def start_timer_for(self, key):
        """start timer to count duration for the specified key (task-case)."""

        if key not in self:
            self[key] = self.__get_initialized_data()

        self[key][self.TIMER] = Timer("counter_timer_%s" % key)

    def start_timer_for_app(self):
        """start timer for the application to count the whole task duration."""

        self._app[self.TIMER] = Timer("counter_timer_%s" % self.MINIMO)

    def stop_timer_for(self, key):
        """stop timer for the specified key (task-case)."""

        self[key][self.TIMER].stop()

    def stop_timer_for_app(self):
        """stop timer for the application."""

        self._app[self.TIMER].stop()

    def get_duration_of(self, key):
        """get timer counted duration of the specified key (task-case)."""

        if key in self and self[key][self.TIMER] is not None:
            return self[key][self.TIMER].duration()
        else:
            return 0

    def get_duration_of_app(self):
        """get timer counted duration of the application."""

        return self._app[self.TIMER].duration()

    def count_flag(self, flag, force_count=False):
        """count total number for the specified flag"""

        if force_count or self._app.get(flag) is None:
            count = 0
            for key, value in self.items():
                count += len(value[flag])

            self._app[flag] = count

        return self._app[flag]

# private
    def __get_initialized_data(self):
        """generate initialized data"""

        return {
            self.ERROR: [],
            self.EXCEPTION: [],
            self.WARNING: [],
            self.SUCCESS: [],
            self.FAILURE: [],
            self.TIMER: None
        }


def make_append_method(flag):
    def _append(self, key, value):
        self.append(key, flag, value)
    return _append


def make_get_method(flag):
    def _get(self, key):
        return self.get(key, flag)
    return _get


def make_count_method(flag):
    def _count(self, force_count=False):
        return self.count_flag(flag, force_count)
    return _count


for flag in (
    Counter.ERROR,
    Counter.EXCEPTION,
    Counter.WARNING,
    Counter.SUCCESS,
    Counter.FAILURE
):
    _append = make_append_method(flag)
    _get = make_get_method(flag)
    _count = make_count_method(flag)
    _name = flag.lower()
    setattr(Counter, "append_%s" % _name, _append)
    setattr(Counter, "get_%s" % _name, _get)
    setattr(Counter, "total_%s" % _name, _count)

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2017-08-08
#
# export:
#   callbacks
#   before_action
#   action_step
#   after_action
#

import inspect
from functools import wraps
from .globals import ctx


class Callbacks(object):
    """object to store defined callbacks.
    """

    def __init__(self):
        super(Callbacks, self).__init__()

        self.before_action_funcs = {}
        self.after_action_funcs = {}
        self.action_step_funcs = {}

    def clear_before_action(self, key):
        """clear before_action filters"""

        self.before_action_funcs[key] = []

    def clear_after_action(self, key):
        """clear after_action filters"""

        self.after_action_funcs[key] = []

    def append_to_before_actions(self, key, value):
        """append function to before_action filters"""

        self._append_to(key, value, self.before_action_funcs)

    def get_before_actions(self, key):
        """get before_action filters, return an empty array if there's
        no function in filter list.
        """

        if key in self.before_action_funcs:
            return self.before_action_funcs[key]
        else:
            return []

    def append_to_action_steps(self, key, value):
        """append function to action_step list"""

        self._append_to(key, value, self.action_step_funcs)

    def get_action_steps(self, key):
        """get action_step list, return an empty array if there's
        no function in that list.
        """

        if key in self.action_step_funcs:
            return self.action_step_funcs[key]
        else:
            return []

    def append_to_after_actions(self, key, value):
        """append function to after_action filters"""

        self._append_to(key, value, self.after_action_funcs)

    def get_after_actions(self, key):
        """get after_action filters, return an empty array if there's
        no function in filter list.
        """

        if key in self.after_action_funcs:
            return self.after_action_funcs[key]
        else:
            return []

    def _append_to(self, key, value, container):
        """append value to container by key"""

        if key in container:
            container[key].append(value)
        else:
            container[key] = [value]


ctx.callbacks = Callbacks()


def _get_caller_id_by_frame(caller):
    return "%s.%s" % (caller.f_locals["__module__"],
                      caller.f_code.co_name)


def before_action(*params):
    """to specify the action will be executed before task."""

    def decorator(f):
        @wraps(f)
        def decorated_func(self, *args, **kwargs):
            return f(self, *args, **kwargs)
        ctx.callbacks.append_to_before_actions(
            _get_caller_id_by_frame(inspect.currentframe().f_back),
            decorated_func)
        return decorated_func
    return decorator


def after_action(*params):
    """to specify the action will be executed after task."""

    def decorator(f):
        @wraps(f)
        def decorated_func(self, *args, **kwargs):
            return f(self, *args, **kwargs)
        ctx.callbacks.append_to_after_actions(
            _get_caller_id_by_frame(inspect.currentframe().f_back),
            decorated_func)
        return decorated_func
    return decorator


def action_step(*params):
    """to specify the action is one task step.
    the specified action should return True or False to
    represent success or failure of that task step.
    """

    def decorator(f):
        @wraps(f)
        def decorated_func(self, *args, **kwargs):
            return f(self, *args, **kwargs)
        ctx.callbacks.append_to_action_steps(
            _get_caller_id_by_frame(inspect.currentframe().f_back),
            decorated_func)
        return decorated_func
    return decorator

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2017-08-08
#


import time


class Timer(object):
    """timer, started at initialized."""

    timers = {}

    def __init__(self, name):
        self.started_at = time.time()
        self.name = name
        self._duration = 0
        self._stopped = False

        Timer.timers[name] = self

    def duration(self):
        """calculate duration started from timer initialized. unit is sec."""

        if not self._stopped:
            self._duration = time.time() - self.started_at

        return self._duration

    def stop(self):
        """stop the timer and count the duration"""

        self._duration = time.time() - self.started_at
        self._stopped = True

    @classmethod
    def get(cls, name):
        """get named timer with {name}"""

        return cls.timers[name]
# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2018-02-25
#


import copy


class AttributeDict(dict):
    """Works like dict but provides methods to get attribute by '.'.

    such as::
        attrDict = AttributeDict({'foo': foo, 'bar': bar})
        print(attrDict.foo)
        attrDict.bar = bar2
    """

    def __init__(self, defaults={}):
        super(AttributeDict, self).__init__(defaults)
        # dict.__init__(self, defaults)

    def __getattr__(self, attrname):
        return self.get(attrname)

    def __setattr__(self, attrname, value):
        self[attrname] = value

    def __deepcopy__(self, memo):
        return AttributeDict(copy.deepcopy(self.copy(), memo))

# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2018-02-26
#

import os
import sys
import time
import click
import runpy
import fnmatch
import collections
from ..helpers import *
from ..globals import *
from multiprocessing.dummy import Pool as ThreadPool
from minimo import __version__


@click.command("version")
def print_version():
    """print minimo version number.

    usage in cli mode:

        $ mmo version

    ----------

    usage in api mode:

        import minimo

        mmo = minimo.Application(
                    interface="api",
                    root_path=instance_project_path)

        # version string
        version = mmo.call("version")
    """

    version = "minimo {}".format(__version__)

    info(version)
    return version


@click.command("ls")
@click.argument("patterns", nargs=-1)
def list_cases(patterns=[]):
    """list all standard task cases.

    usage in cli mode:

        $ mmo ls [pattern...]

    "pattern" supports Unix shell-style wildcards, such as * or ?.
    if not specified "pattern", it will list all standard cases' names under
    "cases" folder. if specified "pattern", it will search the case name by
    "pattern". can give multiple patterns, such as：

        $ mmo ls foo bar*

    tip: can use 'mmo' or 'minimo' as the main command after v0.4.0.

    ----------

    usage in api mode:

        import minimo

        mmo = minimo.Application(
                    interface="api",
                    root_path=instance_project_path)

        # return sorted valid cases
        sorted_cases = mmo.call("ls")
    """

    if ctx.app.inst_path is None:
        error('not in minimo project root folder.')
        return []

    cases = []
    pattern = "|".join([fnmatch.translate(ptn) for ptn in patterns])

    # loop for standard case
    case_dir = os.path.join(ctx.app.inst_path, "cases")
    for _root, _dirs, _files in os.walk(case_dir):
        if "__main__.py" in _files:
            _name = _get_case_name(_root)
            if pattern:
                if re.match(pattern, _name):
                    cases.append(_name)
            else:
                cases.append(_name)

    result = sorted(cases)
    info("\n".join(result))
    stage("totally %d" % len(result))

    return result


@click.command("run")
@click.argument("cases", nargs=-1)
def run_suite(cases):
    """run task suite.

    usage in cli mode:

        $ mmo run [case...]

    can specify some cases separated by whitespace as:

        $ mmo run case1 case2 case3

    and also can specify some suites (case group under one folder) as:

        $ mmo run suite1 suite2 suite3

    minimo will run all cases under those suites.

    tip: can use 'mmo' or 'minimo' as the main command after v0.4.0.

    ----------

    usage in api mode:

        import minimo

        mmo = minimo.Application(
                    interface="api",
                    root_path=instance_project_path)

        # return output file path or None if all failed
        sorted_cases = mmo.call(
                        "run",
                        cases=["suite1", "suite2/case1", suite2/case2])
    """

    if ctx.app.inst_path is None:
        error('not in minimo project root folder.')
        return None

    tasks = collections.OrderedDict()
    task_suite = "task"

    # check case runner
    for case in set(cases):
        runner_path = os.path.join(ctx.app.inst_path, "cases", case)
        task_suite = "{0}_{1}".format(task_suite, case.replace("/", "_"))

        # loop for __main__.py
        valid_case = False
        for _root, _dirs, _files in os.walk(runner_path):
            if "__main__.py" in _files:
                _name = _get_case_name(_root)
                valid_case = True
                if _name not in tasks:
                    tasks[_name] = _root
                    info("add task '%s'" % _name)

        if not valid_case:
            warning(
                "%s is not %s standard case, please run it respectively." % (
                    case, ctx.app.name))
            ctx.counter.append_exception(case, "not standard case")

    # append timestamp
    ctx.suite_name = "{0}_{1}".format(task_suite,
                                      time.strftime("%Y_%m_%d_%H_%M_%S"))

    ctx.output_path = os.path.join(ctx.app.inst_path, "log", ctx.suite_name)

    if "concorrence" == ctx.app.running_type:
        # check out max_thread_count
        if hasattr(ctx.config, "max_thread_count"):
            # up2date minimo version
            max_thread_count = ctx.config.max_thread_count
        elif hasattr(ctx.config, "max_process_count"):
            # previous minimo version
            max_thread_count = ctx.config.max_process_count
        else:
            # no such config, set to 10 threads by default
            max_thread_count = 10

        _run_suite_concurrently(tasks, max_thread_count)
    else:
        _run_suite_serially(tasks)

    return ctx.reporter.report()


def _run_case(case, path, context):
    """run case from the specified path as named 'case'."""

    try:
        stage("running task '%s'..." % case)

        module_path = os.path.abspath(os.path.join(path, ".."))
        if module_path not in sys.path:
            sys.path.insert(0, module_path)

        mainpy_path = os.path.join(path, "__main__.py")
        if os.path.basename(path) != "__main__.py" and \
           os.path.exists(mainpy_path):
            # definitely set the __main__.py path to
            # avoid multiple thread conflict
            _path = mainpy_path
        else:
            _path = path

        runpy.run_path(_path)

    except Exception:
        tb = format_traceback()
        context.counter.append_exception(case, tb)
        error("exception occured while performing '%s':\n%s" % (case, tb))


def _run_suite_serially(tasks):
    """serial type to run cases.

    :param tasks: dict for tasks, key is task name, value is the path for task
                  module, task should have __main__ entry.
    """

    for _name, _path in tasks.items():
        _run_case(_name, _path, ctx)


def _run_suite_concurrently(tasks, max_thread_count=10):
    """concorrence type to run cases.

    :param tasks: dict for tasks, key is task name, value is the path for
                  task module, task should have __main__ entry.

    :param max_thread_count: max thread count to run tasks.
    """

    if not isinstance(max_thread_count, int):
        error("max thread count is not number, please check out your config.")
        return

    if max_thread_count <= 0 or max_thread_count > 1000:
        max_thread_count = 1000

    info("prepare to run suite in %d thread." % max_thread_count)
    pool = ThreadPool(max_thread_count)

    for _name, _path in tasks.items():
        pool.apply_async(
            _run_case,
            kwds={
                "case": _name,
                "path": _path,
                "context": ctx
            })
    pool.close()
    pool.join()


def _get_case_name(case_path):
    """extract case name from case path."""

    return case_path.replace(os.path.join(
        ctx.app.inst_path, "cases"), "")[1:].replace("\\", "/")

# end



# -*- coding:utf-8 -*-


import minimo


class BasePerformer(minimo.Performer):
    """docstring for BasePerformer"""

    def __init__(self, name, logger=None, case_path=None):
        super(BasePerformer, self).__init__(
            name=name, logger=logger, case_path=case_path)


# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2017-08-10
#

import os
import sys
import importlib
import yaml
import logging
from .globals import ctx
from .ext import __autoload__
from .interface import InterfaceFactory
from .helpers import *
from .reporter import Reporter
from .logger import StdoutToFileLogger


class Application(object):
    """the Application object implements the basic entry of
    minimo framework.
    """

    # project name, set in project instance
    name = "minimo"

    # project type, default is "minimo"
    project_type = "minimo"

    # report format type, supported "text", "html" and "xml",
    # default is "html"
    #   text: plain text
    #   html: html web page
    #   xml: xml file, can be recognized by most CI system
    output = "text"

    # case running type, should be "serial" or "concorrence",
    # default is "serial"
    #   serial: run cases one by one
    #   concorrence: run cases concorrently by threads.
    running_type = "serial"

    # redirect stdout message to log file or not.
    # if running in background, there maybe raises exception due to no stdout
    # handler, so set this flag to True, print stdout message to log file.
    # it is forced to redirection under api mode.
    redirect_stdout_to_file = False

    # project modules path, which will be inserted into sys.path at application
    # started. by default, "lib", "ext", "cases", "vendor" will be added
    # mandatory.
    mandatory_modules_path = ["lib", "ext", "cases", "vendor"]
    modules_path = []

    def __init__(self, **attrs):
        super(Application, self).__init__()

        # root path of project instance
        self.inst_path = attrs.pop("root_path", None)

        # project interface, supported "cli", "api", default is "cli"
        #   cli: call commands/functions as command line interface
        #   api: call commands/functions as api
        self.interface = attrs.pop("interface", "cli")

        # flags intializing
        self._loaded_plugins = \
            self._loaded_extensions = \
            self._loaded_modules_path = False

        # load config from yaml under project instance's root_path
        self._load_config()

        self.__interface = InterfaceFactory.get(self.interface, **attrs)

        # add default cli
        for cli in __autoload__:
            self.__interface.add_command(cli)

        ctx.app = self

    def run(self, *args, **kwargs):
        """alias for :meth:`main`."""

        self.main(*args, **kwargs)

    def call(self, *args, **kwargs):
        """alias for :meth:`main`."""

        self.main(*args, **kwargs)

    def main(self, *args, **kwargs):
        """main function entry, we just load modules path to sys.path before
        app running up.
        """

        self._setup()

        result = self.__interface.main(*args, **kwargs)

        self._teardown()

        return result

    def get_command(self, context, name):
        """ We load plugins and extension with built-in commands as these
        should always be the same no matter what the app does.
        """

        self._load_plugins()
        self._load_extensions()

        return self.__interface.get_command(self, context, name)

    def list_commands(self, context):
        """we always load plugins and extensions with minimo
        builtin commands.
        """

        self._load_plugins()
        self._load_extensions()

        rv = set(self.__interface.list_commands(self, context))
        return sorted(rv)

    def is_cli_mode(self):
        return "cli" == self.interface

    def is_api_mode(self):
        return "api" == self.interface

    def echo_to_file(self):
        """check out if print console message to file"""

        return self.redirect_stdout_to_file or self.is_api_mode()

    def echo_to_stdout(self):
        """check out if print console message to stdout"""

        return not self.echo_to_file()

# protected
    def _setup(self):
        """setup runtime environment"""

        # initialize runtime context
        self._init_context()

        # start timer for the application
        ctx.counter.start_timer_for_app()

        # add modules path to sys.path
        self._add_modules_path()

    def _teardown(self):
        """tear down runtime environment"""

        # stop timer for the application
        ctx.counter.stop_timer_for_app()

    def _load_config(self):
        """load project configurations: config.yml"""

        if self.inst_path is not None:
            config_path = os.path.join(self.inst_path, "config.yml")
            if self.inst_path is not None and os.path.exists(config_path):
                with open(config_path, "r") as f:
                    # load yml to config and app attributes
                    cfg = yaml.full_load(f.read())
                    if isinstance(cfg, dict):
                        ctx.config.update(cfg)
                        self.__dict__.update(ctx.config)

    def _init_context(self):
        """initialize runtime context"""

        ctx.reporter = Reporter()

        if self.echo_to_file():
            # redirect stdout to log file
            self._logger = StdoutToFileLogger(
                name="stdout",
                output_path=os.path.join(self.inst_path, "log"),
                outputs={
                    "fulltrace": logging.DEBUG
                }
            )

            sys.stdout = self._logger

    def _add_modules_path(self):
        """walk through modules_path, if there's __init__.py, the folder will
        be added into sys.path.
        """

        if self._loaded_modules_path \
           or self.inst_path is None:
            return

        # add instance path to sys.path
        sys.path.insert(0, self.inst_path)

        for target in set(self.modules_path + self.mandatory_modules_path):
            target_dir = os.path.join(self.inst_path, target)
            if os.path.exists(target_dir):
                for dirpath, dirs, files in os.walk(target_dir):
                    if "__init__.py" in files:
                        sys.path.insert(0, dirpath)

        self._loaded_modules_path = True

    def _load_plugins(self):
        """load plugins which were registered as 'minimo.commands'."""

        if self._loaded_plugins:
            return

        try:
            import pkg_resources
        except ImportError:
            self._loaded_plugins = True
            return

        for ep in pkg_resources.iter_entry_points('minimo.commands'):
            self.__interface.add_command(ep.load(), ep.name)
        self._loaded_plugins = True

    def _load_extensions(self):
        """load extensions for minimo style project. such extensions will be
        placed under project/ext folder, as minimo does, add the extended
        commands to __autoload__ tuple, and minimo will load them
        automatically.
        """

        if self._loaded_extensions:
            return

        if self.inst_path is not None \
           and os.path.exists(os.path.join(self.inst_path, "ext")):
            try:
                for cli in importlib.import_module("ext").__autoload__:
                    self.__interface.add_command(cli)
            except Exception:
                # do nothing
                self._loaded_extensions = True
                return
        self._loaded_extensions = True
# end



# -*- coding:utf-8 -*-
#
# author: philip1134
# date: 2017-08-08
#


import os
import copy
import yaml
from .globals import *
from .helpers import *
from .logger import Logger


class Performer(object):
    """mini-mo base performer to perform task case.

    :attr name: performer name
    :attr app: ref of application instance
    :attr config: deepcopy of project config from config.yml
    """

    def __init__(
        self,
        name,
        logger=None,
        case_path=None
    ):
        """performer constructor

        :param name: task name
        :param logger:
            logging handler, if set to none, will use minimo style logger.

            for customized logger, it should respond to the following methods:
                * open(): setup logger staff
                * close(): teardown logger
                * info(message, *args, **kwargs): print normal information
                                                  message
                * warning(message, *args, **kwargs): print warning message
                * error(message, *args, **kwargs): print error message
                * success(message, *args, **kwargs): print task success message
                * fail(message, *args, **kwargs): print task failure message
        :param case_path: case folder path
        """

        super(Performer, self).__init__()

        self.name = name
        self.app = ctx.app

        if case_path is None:
            self.case_path = None
        else:
            self.case_path = os.path.dirname(case_path)

        self._load_config()

        # setup logger
        if logger is None:
            self.logger = Logger(name=self.name,
                                 output_path=ctx.output_path,
                                 stdout=ctx.app.echo_to_stdout())
        else:
            self.logger = logger

    def run(self):
        """execute performer"""

        self._setup()

        self._do_before_actions() and \
            self._do_action_steps() and \
            self._do_after_actions()

        self._teardown()

# protected
    def _setup(self):
        """setup environment before performer actions"""

        # start timer
        ctx.counter.start_timer_for(self.name)

    def _teardown(self):
        """tear down environment after performer actions"""

        # stop timer
        ctx.counter.stop_timer_for(self.name)

    def _load_config(self):
        """load case configuration under case-folder/config.yml"""

        # copy global config
        self.config = copy.deepcopy(ctx.config)

        # copy case config
        if self.case_path is not None:
            config_path = os.path.join(self.case_path, "config.yml")
            if os.path.exists(config_path):
                with open(config_path, "r") as f:
                    # load yml to config
                    cfg = yaml.full_load(f.read())
                    if isinstance(cfg, dict):
                        self.config.update(cfg)

    def _do_before_actions(self):
        """execute before_action functions"""
        return self._do_actions(
            ctx.callbacks.get_before_actions(self._get_caller_id()))

    def _do_action_steps(self):
        """execute action_step functions"""
        return self._do_actions(
            ctx.callbacks.get_action_steps(self._get_caller_id()), True)

    def _do_after_actions(self):
        """execute after_step functions"""
        return self._do_actions(
            ctx.callbacks.get_after_actions(self._get_caller_id()))

    def _do_actions(self, action_list, action_step=False):
        result = True
        for func in action_list:
            self.logger.info("[%s] start action '%s'..." % (
                self.name, func.__name__))

            try:
                r = func(self)
            except Exception:
                r = False
                tb = format_traceback()
                ctx.counter.append_exception(self.name, tb)
                self.logger.error(
                    '[%s] error occured, please check out your task code!'
                    '\n%s' % (self.name, tb))

            if isinstance(r, bool):
                _r = r
            else:
                _r = True

            if action_step:
                if _r:
                    self.logger.success("[%s] succeed to perform '%s'" % (
                                        self.name,
                                        func.__name__))
                else:
                    self.logger.fail("[%s] fail to perform '%s'" % (
                                     self.name,
                                     func.__name__))
            elif not _r:
                result = False
                self.logger.fail("[%s] error occured while performing '%s'" % (
                                 self.name,
                                 func.__name__))
                break

        return result

    def _get_caller_id(self):
        return "%s.%s" % (self.__module__, self.__class__.__name__)

# end
